#!/usr/bin/python3

import collections
import itertools
import operator
from typing import Iterable, NamedTuple

import itanium_demangler as demangler
from cle.backends.symbol import Symbol
from cle.memory import Clemory

import angr

from .. import demangler_helpers as dh


def read_cstring(
    mem: Clemory, addr: int, encoding: str = "utf-8", chunk: int = 16, **kwargs
) -> str:
    """unpacks a variable-length, zero-terminated string from the binary"""
    try:
        unpacker = itertools.takewhile(
            lambda x: x > 0,
            itertools.chain.from_iterable(
                mem.load(addr + o, chunk) for o in itertools.count(0, chunk)
            ),
        )
        return bytes(unpacker).decode(encoding, **kwargs)
    except Exception:
        pass
    return ""


class VtableDisambiguator(angr.Analysis):
    syms_by_addr: dict[int, set[Symbol]]

    # given a vtable symbol, returns childrens' vtable symbols in an undefined order
    subclass_map: dict[Symbol, set[Symbol]]

    # given a vtable symbol, returns the parents' vtable symbols in ascending order
    superclass_map: dict[Symbol, list[Symbol]]

    # we use this to test if a function belongs to a class with a vtable
    # this way we can prune static functions from the candidate pool
    vtable_instances: set[demangler.Node]

    def __init__(self):
        self.loader = self.project.loader
        self.memory = self.loader.memory
        self.analyze()

    def analyze(self) -> None:
        self.syms_by_addr = collections.defaultdict(set)
        vtable_syms = set()
        self.vtable_instances = set()
        for symbol in self.loader.symbols:
            self.syms_by_addr[symbol.rebased_addr].add(symbol)
            if symbol.name.startswith("_ZTV") and not symbol.name.startswith("_ZTVZ"):
                vtable_syms.add(symbol)
                self.vtable_instances.add(
                    dh.extract_vtable_typename(demangler.parse(symbol.name))
                )

        # we associate a type with their subclasses so we can check them to disambiguate duplicates on the base
        self.subclass_map = collections.defaultdict(set)
        self.superclass_map = collections.defaultdict(list)

        for vtsym in vtable_syms:
            if vtsym.is_import:
                continue

            vt_typeinfo = self.memory.unpack_word(vtsym.rebased_addr + 0x4)
            for typeinfo_ptr, typeinfo_name in self.dump_class_parents(vt_typeinfo):
                svtsym = self.loader.find_symbol(f"_ZTV{typeinfo_name}")
                if not svtsym:
                    continue
                self.subclass_map[svtsym].add(vtsym)
                self.superclass_map[vtsym].append(svtsym)

    def dump_class_parents(self, typeinfo_ptr: int) -> Iterable[tuple[int, str]]:
        """
        Returns a list of mangled typenames in ascending order (towards base classes at the end).
        This takes a pointer to the class's typeinfo structure.
        """
        # HACK: adjust offset by -8h for angr to resolve to the correct symbol
        #       IDA indicates the offset is at sym+8h
        #       lief seems to resolve this correctly without offsetting by -8h
        if not typeinfo_ptr:
            return
        typeinfo_class_sym = self.loader.find_symbol(
            self.memory.unpack_word(typeinfo_ptr) - 0x8
        )
        if not typeinfo_class_sym:
            return

        typeinfo_name = read_cstring(self.memory, self.memory.unpack_word(typeinfo_ptr + 0x04))
        if typeinfo_class_sym.name == "_ZTVN10__cxxabiv120__si_class_type_infoE":
            # single inheritance
            nested_typeinfo_ptr = self.memory.unpack_word(typeinfo_ptr + 0x08)
            yield (typeinfo_ptr, typeinfo_name)
            yield from self.dump_class_parents(nested_typeinfo_ptr)
        elif typeinfo_class_sym.name == "_ZTVN10__cxxabiv121__vmi_class_type_infoE":
            # multiple inheritance
            nested_typeinfo_ptr = self.memory.unpack_word(typeinfo_ptr + 0x10)
            yield (typeinfo_ptr, typeinfo_name)
            yield from self.dump_class_parents(nested_typeinfo_ptr)
        elif typeinfo_class_sym.name == "_ZTVN10__cxxabiv117__class_type_infoE":
            yield (typeinfo_ptr, typeinfo_name)
        else:
            raise ValueError("unknown typeinfo class", typeinfo_class_sym)

    def resolve_ambiguous_vfn(
        self, vtidx: int, ambig_fnsyms: set[Symbol], related_vtsyms: Iterable[Symbol]
    ) -> Symbol | None:
        """
        Resolves an ambiguous virtual function by attempting to match against symbols in related
        classes' vtables.
        """
        # ambiguous function
        # walk through vtables of subclasses to find matching prototype at the given position
        # if unsuccessful, walk through parent classes containing the vtable
        # returns the exact name of the symbol used for this function
        funcsig_set = set()
        for fnsym in ambig_fnsyms:
            funcsig_set.add(dh.extract_method_signature(demangler.parse(fnsym.name)))

        non_class_syms = set()
        for sym in ambig_fnsyms:
            fnname = dh.extract_function_name(demangler.parse(sym.name))
            if not any(vtname == fnname[: len(vtname)] for vtname in self.vtable_instances):
                non_class_syms.add(sym)

        if len(ambig_fnsyms - non_class_syms) == 1:
            # all but one of the ambiguous symbols are not potential virtual class methods
            return (ambig_fnsyms - non_class_syms).pop()

        if len(funcsig_set) == 1:
            # we only have one name-signature combination here
            return set(ambig_fnsyms).pop()

        rebased_fnsym_addrs = set(fnsym.rebased_addr for fnsym in ambig_fnsyms)

        for fnsym in ambig_fnsyms - non_class_syms:
            funcsig = dh.extract_method_signature(demangler.parse(fnsym.name))
            for svt in related_vtsyms:
                # only check functions at the same index
                subfn_addr = self.loader.fast_memory_load_pointer(
                    svt.rebased_addr + 0x4 * (2 + vtidx)
                )

                if subfn_addr in rebased_fnsym_addrs:
                    # skip vtable entries that call the same exact function
                    continue

                subfns = self.syms_by_addr.get(subfn_addr)
                if not subfns:
                    continue

                for subfn in subfns:
                    sub_funcsig = dh.extract_method_signature(demangler.parse(subfn.name))
                    if funcsig != sub_funcsig:
                        continue

                    # we found a subclass with a name / prototype that matches the base class
                    return fnsym
        return None

    def get_possible_vtable_set_candidates(
        self, vtsym: Symbol, vtidx: int
    ) -> Iterable[set[Symbol]]:
        # It's possible that this class inherited a base version of a method that is specialized
        # in a different part of the class hierarchy, so find the base implementor then return
        # it along with its descendants.
        psym = self.get_parent_vt_implementing(vtsym, vtidx)
        yield {psym} | self.subclass_map[psym]

    def get_parent_vt_implementing(self, vtsym: Symbol, vtidx: int) -> Symbol:
        # walk up parents until we reach the first one to implement this method
        for super_vtsym, gp_vtsym in itertools.pairwise(self.superclass_map[vtsym]):
            vt_first, *_ = self.get_vfptrs_from_table(gp_vtsym)
            if len(vt_first) < vtidx:
                return super_vtsym
        return self.superclass_map[vtsym][-1]

    def get_vfptrs_from_table(self, vtsym: Symbol) -> list[list[int]]:
        # returns a list of addresses for each vtable present on the class
        class VTableFunction(NamedTuple):
            tblidx: int
            address: int

        table_index = 0
        function_list: list[VTableFunction] = []

        # we include the initial upcast offset and class typeinfo here
        vtable_range = enumerate(range(vtsym.rebased_addr, vtsym.rebased_addr + vtsym.size, 4))

        for n, addr in vtable_range:
            # get symbols that map to that address
            # NOTE: classes that don't implement virtual destructors have nullptrs
            deref = self.loader.fast_memory_load_pointer(addr)
            fnsyms = set(self.syms_by_addr.get(deref) or set()) if deref else set()
            if not fnsyms:
                # vtable boundary; consume typeinfo so it doesn't get added to the list
                # NOTE: we don't actually care if the indices skip; all that matters is that the
                # functions are grouped correctly
                table_index += 1
                next(vtable_range)

                for n1, addr1 in vtable_range:
                    # consume any leading nullptrs (destructors that were optimized out)
                    # this also consumes the next actual function pointer
                    # we could probably get away with looking for it in the parent class...
                    deref = self.loader.fast_memory_load_pointer(addr1)
                    function_list.append(VTableFunction(table_index, deref))
                    if deref != 0:
                        break
                continue
            function_list.append(VTableFunction(table_index, deref))

        return [
            list(vfn.address for vfn in vtbl)
            for tblidx, vtbl in itertools.groupby(
                function_list, key=operator.attrgetter("tblidx")
            )
        ]


angr.analyses.AnalysesHub.register_default("VtableDisambiguator", VtableDisambiguator)
