#!/usr/bin/python3

# helper functions to extract information from a demangler AST
#
# the returned tuples are not well-defined, but is expected to be internally consistent
# i.e. signatures from different classes should match

import itanium_demangler as demangler


def extract_method_classname(node: demangler.Node) -> tuple[demangler.Node, ...]:
    def _extract(node: demangler.Node):
        match node:
            case node if node.kind == "qual_name":
                return node.value[:-1]
        raise ValueError(f"Unexpected node {node!r}")

    if node.kind == "func":
        return _extract(node.name)

    raise ValueError(f"{node} is not a function")


def extract_function_name(node: demangler.Node) -> tuple[demangler.Node, ...]:
    # this name will include the class at the start
    def _extract(node: demangler.Node):
        match node:
            case node if node.kind in "name":
                return node.value
            case node if node.kind == "cv_qual":
                return _extract(node.value)
            case node if node.kind == "qual_name":
                return node.value
        raise ValueError(f"Unexpected node {node!r}")

    if node.kind == "func":
        return _extract(node.name)

    raise ValueError(f"{node} is not a function")


def extract_vtable_typename(node: demangler.Node):
    # extracts the typename of a vtable, guaranteed as a tuple of nodes
    def _extract(node: demangler.Node):
        match node:
            case node if node.kind == "qual_name":
                return node.value
            case node if node.kind == "name":
                return (node,)
        raise ValueError(f"Unexpected node {node!r}")

    if node.kind == "vtable":
        return _extract(node.value)

    raise ValueError(f"{node} is not a vtable")


def extract_method_fname(node: demangler.Node) -> tuple[demangler.Node, ...]:
    # returns method name with any associated qualifiers
    def _extract(node: demangler.Node):
        match node:
            case node if node.kind == "cv_qual":
                return _extract(node.value)
            case node if node.kind == "qual_name":
                return (node.value[-1],)
        raise ValueError(f"Unexpected node {node!r}")

    if node.kind == "func":
        return _extract(node.name)
    elif node.kind == "nonvirt_thunk":
        return _extract(node.value.name)

    raise ValueError(f"{node} is not a function")


def extract_method_signature(node: demangler.Node) -> tuple[demangler.Node, ...]:
    # returns tuples describing the method name, argument types, and return type
    # this is expected to strip the class name
    if node.kind == "func":
        return (extract_method_fname(node), node.arg_tys, node.ret_ty)
    elif node.kind == "nonvirt_thunk":
        return extract_method_signature(node.value)
    raise ValueError(f"{node} is not a function")
