#!/usr/bin/python3

import ast
import re


class ByteSignature:
    """
    A sequence of hex bytes to be searched.
    """

    pattern: list[str]
    expr: re.Pattern

    def __init__(self, pattern: str):
        self.pattern = pattern.split()

        # creates escaped byte pattern with hex literals or wildcard as appropriate
        self.expr = re.compile(
            b"".join(
                [re.escape(bytes([int(b, 16)])) if b != "??" else b"." for b in self.pattern]
            ),
            re.S,
        )

    @property
    def display_str(self):
        # render as a bracketed, space-delimited hex string where wildcard bytes are denoted with '??'
        return f"[{' '.join(self.pattern)}]".lower()

    @property
    def gameconf_str(self):
        # render as a SourceMod-style escaped string
        return "".join(r"\x" + b.upper() if b != "??" else r"\x2A" for b in self.pattern)

    @property
    def length(self):
        return len(self.pattern)

    def __repr__(self):
        return f"ByteSignature({self.display_str})"


class Code:
    """
    A class to parse and execute Python code in a reduced (but not secure) environment.
    """

    code_ast: ast.AST

    def __init__(self, v: str):
        self.code_ast = ast.parse(v, mode="eval")

    def eval(self, **locals):
        code = compile(self.code_ast, "<CONFIG>", mode="eval")
        return eval(code, None, locals)

    def __str__(self):
        return ast.unparse(self.code_ast)


class IntLiteral(int):
    """
    Subclass of ``int`` that performs implicit conversion of integer values.  This allows
    configuration using numeric literals in non-base-10 configurations, such as 0xAA.
    """

    def __new__(cls, *args, **kwargs):
        return super(IntLiteral, cls).__new__(cls, *args, base=0)
