#!/usr/bin/python3

# vtable helpers
# this should probably be cleaned up and moved somewhere else

import collections
import itertools
import operator
import typing

import itanium_demangler as demangler
from cle.backends.symbol import Symbol

from . import demangler_helpers as dh

if typing.TYPE_CHECKING:
    from .validate import LinuxBinary

VTable = list[Symbol]

# hotfix for demangler
demangler._is_ctor_or_dtor = demangler.is_ctor_or_dtor


def reorder_vfns_windows_estimate(symbols: list[Symbol], start_pos: int) -> list[Symbol]:
    # reorders a given subclass-level slice of linux symbols to reflect windows ordering
    name_buckets = collections.defaultdict(list)
    for n, symbol in enumerate(symbols):
        # collect overrides into buckets based on function name
        dmsym = demangler.parse(symbol.name)
        if dmsym:
            name_buckets[dh.extract_method_fname(dmsym)].append(symbol)
        else:
            # HACK: preserves positions of references to __cxa_pure_virtual
            name_buckets[(dmsym, n)].append(symbol)

    output_symbols: list[Symbol] = []
    for fname, syms in name_buckets.items():
        # on windows, overloads are made consecutive and in reverse of declared order
        # iteration order is guaranteed as of py3.7+ to be the insertion order,
        # so this should output symbols otherwise in their original order
        output_symbols.extend(reversed(syms))

    return output_symbols


def get_windows_vtables_from(bin: "LinuxBinary", vt: Symbol) -> VTable:
    vtda = bin.vtable_disambiguator
    vt_typeinfo = bin.angr.loader.memory.unpack_word(vt.rebased_addr + 0x4)

    vt_parent_spans = [1]
    for typeinfo_ptr, name in reversed(list(vtda.dump_class_parents(vt_typeinfo))):
        vt_parent = bin.angr.loader.find_symbol(f"_ZTV{name}")
        if not vt_parent:
            continue
        vt_parent_first, *_ = vtda.get_vfptrs_from_table(vt_parent)
        vt_parent_spans.append(len(vt_parent_first))

    vt_first, *vt_others = get_vtables_from_address(bin, vt)
    thunk_fns = set()
    for vt_other in vt_others:
        for sym in vt_other:
            dmsym = demangler.parse(sym.name)
            if dmsym and dmsym.kind == "nonvirt_thunk":
                thunk_fns.add(dh.extract_method_signature(dmsym))

    vt_out = []
    for vt_low, vt_high in itertools.pairwise(vt_parent_spans):
        # we can only reorder overloads within the class they were initially specified
        # e.g. CTFPlayer's ChangeTeam cannot be merged with CBaseEntity's
        class_vfns = []

        for sym in vt_first[vt_low:vt_high]:
            # filter MI thunks
            dmsym = demangler.parse(sym.name)
            if dmsym:
                if (
                    not demangler.is_ctor_or_dtor(dmsym)
                    and dh.extract_method_signature(dmsym) in thunk_fns
                ):
                    continue
            else:
                # __cxa_pure_virtual returns None here; we still add it to the vtable slice
                pass
            class_vfns.append(sym)
        vt_out.extend(reorder_vfns_windows_estimate(class_vfns, vt_low))

    return vt_out


def get_constrained_vfn(
    bin: "LinuxBinary", vt: Symbol, vfnidx: int, candidate_fnsyms: set[Symbol]
) -> Symbol | None:
    for psym in bin.vtable_disambiguator.superclass_map[vt]:
        if psym.name not in bin.vtable_constraint:
            continue
        vt_const = bin.vtable_constraint[psym.name]
        for fnsym in candidate_fnsyms:
            if fnsym.name in vt_const and vt_const[fnsym.name] == vfnidx:
                return fnsym
    return None


def get_vtables_from_address(bin: "LinuxBinary", vt: Symbol) -> list[VTable]:
    vtda = bin.vtable_disambiguator

    # returns a list of vtables for each vtable present on the class
    class VTableFunction(typing.NamedTuple):
        tblidx: int
        possible_syms: set[Symbol]

    function_list: list[VTableFunction] = []
    vptr_lists = vtda.get_vfptrs_from_table(vt)
    for table_index, vptrs in enumerate(vptr_lists):
        for n, vptr in enumerate(vptrs):
            # get symbols that map to that address
            if vptr == 0:
                # HACK: some virtual destructors got optimized out and are represented by nullptrs
                # TODO: it might be better to pull destructors from the parent
                call_unexpected = bin.angr.loader.find_symbol("__cxa_call_unexpected")
                assert call_unexpected
                function_list.append(VTableFunction(table_index, {call_unexpected}))
                continue

            fnsyms = set(vtda.syms_by_addr.get(vptr) or set()) if vptr else set()

            if len(fnsyms) == 1:
                function_list.append(VTableFunction(table_index, fnsyms))
                continue
            elif len(fnsyms) > 1:
                constrained_sym = get_constrained_vfn(bin, vt, n, fnsyms)
                if constrained_sym:
                    function_list.append(VTableFunction(table_index, {constrained_sym}))
                    continue

                # function in vtable is referenced by multiple names; perform disambiguation
                matched_overload = None
                for related in vtda.get_possible_vtable_set_candidates(vt, n):
                    matched_overload = vtda.resolve_ambiguous_vfn(n, fnsyms, related)
                    if matched_overload:
                        break

                # it's possible that the other function(s) is/are resolveable.
                # without doing multiple passes and saving the disambiguity somewhere it'll be difficult to match

                if matched_overload:
                    function_list.append(VTableFunction(table_index, {matched_overload}))
                    continue

            function_list.append(VTableFunction(table_index, fnsyms))

    for n, vfn in enumerate(function_list):
        if n == 0:
            # HACK: skip duplicated references to destructor
            # we should be doing this at the disambiguation stage
            continue
        elif len(vfn.possible_syms) == 1:
            continue

        # we should never receive an empty ``VTableFunction.possible_syms``
        # for now we need to assert that a function address is unambiguous given the context
        vt_name = demangler.parse(vt.name)
        candidate_names = set(sym.name for sym in vfn.possible_syms)
        raise Exception(f"Ambiguity in {vt_name} position {n}; candidates {candidate_names}")

    return [
        list(vfn.possible_syms.pop() for vfn in vtbl)
        for tblidx, vtbl in itertools.groupby(function_list, key=operator.attrgetter("tblidx"))
    ]
